# Copyright 2018-2022 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import api_timeout, gpi, info, measure, program, record, scan, \
    set_parameter, statistics, threads

__all__ = [api_timeout, gpi, info, measure, program, record, scan,
           set_parameter, statistics, threads]
"""This list of available command modules.  Each module must contain a 
parser_config(subparser) function.  The function must return the callable(args)
that will be executed for the command."""



# Copyright 2022 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pyjoulescope_driver import Driver, __version__
import numpy as np
import os
import platform
import psutil
import sys


def parser_config(p):
    """Joulescope info."""
    p.add_argument('--verbose', '-v',
                   action='store_true',
                   help='Display verbose information.')
    p.add_argument('--open', '-o',
                   choices=['defaults', 'restore'],
                   default='defaults',
                   help='The device open mode.  Defaults to "defaults".')
    p.add_argument('device_path',
                   nargs='*',
                   help='The target device for this command. ' +
                        'Provide "*" to show details for all connected devices.')
    return on_cmd


def version_to_str(version):
    if isinstance(version, str):
        return version
    v_patch = version & 0xffff
    v_minor = (version >> 16) & 0xff
    v_major = (version >> 24) & 0xff
    return f'{v_major}.{v_minor}.{v_patch}'


def _sys_info():
    try:
        from pyjls import __version__ as jls_version
    except ImportError:
        jls_version = 'uninstalled'
    try:
        os.environ['JOULESCOPE_BACKEND'] = 'none'
        from joulescope import __version__ as joulescope_version
    except ImportError:
        joulescope_version = 'uninstalled'

    cpufreq = psutil.cpu_freq()
    vm = psutil.virtual_memory()
    vm_available = (vm.total - vm.used) / (1024 ** 3)
    vm_total = vm.total / (1024 ** 3)
    return f"""\

    SYSTEM INFORMATION
    ------------------
    python               {sys.version}
    python impl          {platform.python_implementation()}
    platform             {platform.platform()}
    processor            {platform.processor()}
    CPU cores            {psutil.cpu_count(logical=False)} physical, {psutil.cpu_count(logical=True)} total
    CPU frequency        {cpufreq.current:.0f} MHz ({cpufreq.min:.0f} MHz min to {cpufreq.max:.0f} MHz max)   
    RAM                  {vm_available:.1f} GB available, {vm_total:.1f} GB total ({vm_available/vm_total *100:.1f}%)
    
    PYTHON PACKAGE INFORMATION
    --------------------------
    jls                  {jls_version}
    joulescope           {joulescope_version}
    numpy                {np.__version__}
    pyjoulescope_driver  {__version__}
    """


_JOULESCOPE_INFORMATION = """
    JOULESCOPE INFORMATION
    ----------------------"""


def _list_devices(driver):
    txt = []
    device_paths = driver.device_paths()
    if len(device_paths) is None:
        txt.append('No connected Joulescopes found')
    else:
        for device_path in device_paths:
            try:
                driver.open(device_path, mode='restore')
            except Exception:
                txt.append(f'    {device_path}: could not open')
                continue
            try:
                if '/js220/' in device_path:
                    fw = version_to_str(driver.query(f'{device_path}/c/fw/version'))
                    hw = version_to_str(driver.query(f'{device_path}/c/hw/version'))
                    fpga = version_to_str(driver.query(f'{device_path}/s/fpga/version'))
                    txt.append(f'    {device_path}: hw={hw}, fw={fw}, fpga={fpga}')
                else:
                    txt.append(f'    {device_path}')
            except Exception:
                txt.append(f'    {device_path}: could not retrieve details')
            finally:
                driver.close(device_path)
    txt.append('')
    return '\n'.join(txt)


class Info:

    def __init__(self, device_paths):
        self._device_paths = device_paths
        self._meta = {}
        self._values = {}

    def _on_pub(self, topic, value):
        self._values[topic] = value

    def _on_metadata(self, topic, value):
        if topic[-1] == '$':
            topic = topic[:-1]
        self._meta[topic] = value

    def run(self, args):
        if not self._device_paths:
            print(_sys_info())
            print(_JOULESCOPE_INFORMATION)
            with Driver() as d:
                print(_list_devices(d))
                return 0

        with Driver() as d:
            d.log_level = args.jsdrv_log_level
            devices = d.device_paths()
            if '*' in self._device_paths:
                self._device_paths = devices
            for device_path in self._device_paths:
                if device_path not in devices:
                    print(f'{device_path} requested but not found')
                    continue
                self._meta.clear()
                self._values.clear()
                d.open(device_path, mode=args.open)
                fn = self._on_metadata  # use same bound method for unsubscribe
                d.subscribe(device_path, 'metadata_rsp_retain', fn)
                d.unsubscribe(device_path, fn)
                fn = self._on_pub  # use same bound method for unsubscribe
                d.subscribe(device_path, 'pub_retain', fn)
                d.unsubscribe(device_path, fn)
                if args.verbose:
                    print(f'{device_path} metadata:')
                    for key, value in self._meta.items():
                        subtopic = key[len(device_path) + 1:]
                        print(f'  {subtopic} {value}')
                print(f'{device_path} values:')
                for key, value in self._values.items():
                    meta = self._meta.get(key, None)
                    if meta is not None:
                        fmt = meta.get('format', None)
                        if fmt == 'version':
                            value = version_to_str(value)
                    subtopic = key[len(device_path) + 1:]
                    print(f'  {subtopic} = {value}')
        return 0


def on_cmd(args):
    return Info(args.device_path).run(args)



# Copyright 2022 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__version__ = "1.6.0"

__title__ = "pyjoulescope_driver"
__description__ = 'Joulescope™ driver'
__url__ = 'https://joulescope.readthedocs.io'
__author__ = 'Jetperch LLC'
__author_email__ = 'joulescope-dev@jetperch.com'
__license__ = 'Apache 2.0'
__copyright__ = 'Copyright 2022-2023 Jetperch LLC'

__all__ = ['__version__', '__title__', '__description__', '__url__',
           '__author__', '__author_email__', '__license__',
           '__copyright__']



# Copyright 2023 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pyjoulescope_driver import Driver, time64
import sys
import time


def parser_config(p):
    """Measure energy and charge over the specified duration."""
    p.add_argument('--duration',
                   default=1.0,
                   type=time64.duration_to_seconds,
                   help='The capture duration in float seconds. '
                        + 'Add a suffix for other units: s=seconds, m=minutes, h=hours, d=days')
    p.add_argument('--no-progress',
                   action='store_const',
                   default=_on_progress,
                   const=None,
                   help='Skip progress display.')
    p.add_argument('--show-only',
                   choices=['energy', 'charge'],
                   help='Show only the result in SI units, use with --no-progress')
    return on_cmd


def _on_progress(fract, message):
    # The MIT License (MIT)
    # Copyright (c) 2016 Vladimir Ignatev
    #
    # Permission is hereby granted, free of charge, to any person obtaining
    # a copy of this software and associated documentation files (the "Software"),
    # to deal in the Software without restriction, including without limitation
    # the rights to use, copy, modify, merge, publish, distribute, sublicense,
    # and/or sell copies of the Software, and to permit persons to whom the Software
    # is furnished to do so, subject to the following conditions:
    #
    # The above copyright notice and this permission notice shall be included
    # in all copies or substantial portions of the Software.
    #
    # THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
    # INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
    # PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
    # FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
    # OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
    # OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
    fract = min(max(float(fract), 0.0), 1.0)
    bar_len = 25
    filled_len = int(round(bar_len * fract))
    percents = int(round(100.0 * fract))
    bar = '=' * filled_len + '-' * (bar_len - filled_len)

    msg = f'[{bar}] {percents:3d}% {message:40s}\r'
    sys.stdout.write(msg)
    sys.stdout.flush()


def measure(driver, device, duration=None, on_progress=None):
    """Measure energy over a duration.

    :param driver: The driver instance.
    :param device: The device topic prefix.
    :param duration: The duration in seconds.
    :param on_progress: The callable(fract, message) for progress.
    :return: dict containing keys energy, charge, duration.
    """
    duration = 30 if duration is None else int(duration)
    data = []
    if duration <= 0:
        return 0, 0
    total_count = 10 * duration + 1  # need one extra since use first as a baseline

    def _on_statistics_value(topic, value):
        data.append(value)

    driver.subscribe(device + '/s/stats/value', 'pub', _on_statistics_value)
    try:
        while len(data) < total_count:
            time.sleep(0.050)
            duration_remaining = (total_count - len(data)) / 10
            if callable(on_progress):
                on_progress(len(data) / total_count, f' {duration_remaining:.1f} seconds left')
    except KeyboardInterrupt:
        pass
    finally:
        driver.unsubscribe(device + '/s/stats/value', _on_statistics_value)

    if len(data) > total_count:
        data = data[:total_count]
    return {
        'energy': data[-1]['accumulators']['energy']['value'] - data[0]['accumulators']['energy']['value'],
        'charge': data[-1]['accumulators']['charge']['value'] - data[0]['accumulators']['charge']['value'],
        'duration': (len(data) - 1) / 10.0,
    }


def on_cmd(args):
    with Driver() as d:
        devices = d.device_paths()
        if len(devices) != 1:
            print('Found %d devices', len(devices))
            return 1
        device = devices[0]
        d.open(device)
        d.publish(device + '/s/i/range/mode', 'auto')
        d.publish(device + '/s/stats/scnt', 100_000)  # 10 Hz update rate
        d.publish(device + '/s/stats/ctrl', 1)
        data = measure(d, device, duration=args.duration, on_progress=args.no_progress)
        d.publish(device + '/s/stats/ctrl', 0)
        d.close(device)

    if args.no_progress is not None:
        sys.stdout.write('                                                            \r')
        sys.stdout.flush()
    if args.show_only == 'energy':
        print(data['energy'])
    elif args.show_only == 'charge':
        print(data['charge'])
    else:
        print(f"energy={data['energy']} J")
        print(f"charge={data['charge']} C")
        print(f"duration={data['duration']} s")
        print(f"current_average={data['charge'] / data['duration']} A")
        print(f"power_average={data['energy'] / data['duration']} W")
    return 0



#!/usr/bin/env python3
# Copyright 2018-2023 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Joulescope driver python setuptools module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""


# Always prefer setuptools over distutils
import setuptools
from setuptools import Command
from setuptools.errors import ExecError
from setuptools.command.sdist import sdist
import numpy as np
import os
import platform
import subprocess
import sys


MYPATH = os.path.dirname(os.path.abspath(__file__))
VERSION_PATH = os.path.join(MYPATH, 'pyjoulescope_driver', 'version.py')
C_INCS = [
    os.path.join(MYPATH, 'include'),
    os.path.join(MYPATH, 'include_private'),
    os.path.join(MYPATH, 'third-party', 'tinyprintf'),
    np.get_include(),
]

try:
    from Cython.Build import cythonize
    USE_CYTHON = os.path.isfile(os.path.join(MYPATH, 'pyjoulescope_driver', 'binding.pyx'))
except ImportError:
    USE_CYTHON = False


about = {}
with open(VERSION_PATH, 'r', encoding='utf-8') as f:
    exec(f.read(), about)

extra_compile_args = []
extra_link_args = []
posix_sources = [
    'src/backend/posix.c',
    'src/backend/libusb/backend.c',
    'src/backend/libusb/msg_queue.c',
    'third-party/libusb/libusb/core.c',
    'third-party/libusb/libusb/descriptor.c',
    'third-party/libusb/libusb/hotplug.c',
    'third-party/libusb/libusb/io.c',
    'third-party/libusb/libusb/strerror.c',
    'third-party/libusb/libusb/sync.c',
    'third-party/libusb/libusb/os/events_posix.c',
    'third-party/libusb/libusb/os/threads_posix.c',
]


if platform.system() == 'Windows':
    sources = [
        'src/backend/winusb/backend.c',
        'src/backend/winusb/device_change_notifier.c',
        'src/backend/winusb/msg_queue.c',
        'src/backend/windows.c',
    ]
    libraries = ['Setupapi', 'Winusb', 'user32', 'winmm']
    extra_compile_args = []
elif 'armv7' in platform.machine():
    sources = posix_sources
    libraries = ['pthread', 'm']
    extra_compile_args = []
elif platform.processor() == 'aarch64' or platform.machine() == 'arm64':
    sources = posix_sources
    libraries = ['pthread', 'm']
    extra_compile_args = []
else:
    sources = posix_sources
    libraries = ['pthread', 'm']
    extra_compile_args = []


if platform.system() == 'Darwin':
    sources += [
        'third-party/libusb/libusb/os/darwin_usb.c'
    ]
    C_INCS.extend([
        os.path.join(MYPATH, 'third-party/libusb/libusb'),
        os.path.join(MYPATH, 'third-party/libusb/include/macos'),
    ])
    libraries += ['objc']
    extra_link_args = [
                  '-framework', 'IOKit',
                  '-framework', 'CoreFoundation',
                  '-framework', 'Security']
elif platform.system() == 'Linux':
    sources += [
        'third-party/libusb/libusb/os/linux_udev.c',
        'third-party/libusb/libusb/os/linux_usbfs.c',
    ]
    C_INCS.extend([
        os.path.join(MYPATH, 'third-party/libusb/libusb'),
        os.path.join(MYPATH, 'third-party/libusb/include/linux'),
    ])
    libraries += ['udev']

ext = '.pyx' if USE_CYTHON else '.c'
extensions = [
    setuptools.Extension('pyjoulescope_driver.binding',
                         sources=[
                                     'pyjoulescope_driver/binding' + ext,
                                     'src/buffer.c',
                                     'src/buffer_signal.c',
                                     'src/calibration_hash.c',
                                     'src/cstr.c',
                                     'src/devices.c',
                                     'src/downsample.c',
                                     #'src/emu.c',
                                     #'src/emulated.c',
                                     'src/error_code.c',
                                     'src/js110_cal.c',
                                     'src/js110_sample_processor.c',
                                     'src/js110_stats.c',
                                     'src/js110_usb.c',
                                     'src/js220_i128.c',
                                     'src/js220_params.c',
                                     'src/js220_usb.c',
                                     'src/js220_stats.c',
                                     'src/jsdrv.c',
                                     'src/json.c',
                                     'src/log.c',
                                     'src/pubsub.c',
                                     'src/meta.c',
                                     'src/sample_buffer_f32.c',
                                     'src/statistics.c',
                                     'src/time.c',
                                     'src/time_map_filter.c',
                                     'src/topic.c',
                                     'src/union.c',
                                     'src/version.c',
                                     'third-party/tinyprintf/tinyprintf.c',
                                     ] + sources,
                         include_dirs=C_INCS,
                         libraries=libraries,
                         extra_compile_args=[] + extra_compile_args,
                         extra_link_args=extra_link_args,
                         ),
]

if USE_CYTHON:
    from Cython.Build import cythonize
    extensions = cythonize(extensions, compiler_directives={'language_level': '3'})  # , annotate=True)


# Get the long description from the README file
with open(os.path.join(MYPATH, 'README.md'), 'r', encoding='utf-8') as f:
    long_description = f.read()


class BuildSdistCommand(sdist):
    """Custom sdist build command."""

    def run(self):
        cmd = os.path.join(MYPATH, 'pyjoulescope_driver', 'release.py')
        subprocess.run([sys.executable, cmd])
        sdist.run(self)


class CustomBuildDocs(Command):
    """Custom command to build docs locally."""

    description = 'Build docs.'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        # sphinx-build -b html docs build\docs_html
        # defer import so not all setups require sphinx
        from sphinx.application import Sphinx
        from sphinx.util.console import nocolor, color_terminal
        nocolor()
        source_dir = os.path.join(MYPATH, 'doc', 'sphinx')
        target_dir = os.path.join(MYPATH, 'build', 'docs_html')
        doctree_dir = os.path.join(target_dir, '.doctree')
        app = Sphinx(source_dir, source_dir, target_dir, doctree_dir, 'html')
        app.build()
        if app.statuscode:
            raise ExecError(
                'caused by %s builder.' % app.builder.name)


setuptools.setup(
    name=about['__title__'],
    version=about['__version__'],
    description=about['__description__'],
    long_description=long_description,
    long_description_content_type='text/markdown',
    url=about['__url__'],
    author=about['__author__'],
    author_email=about['__author_email__'],
    license=about['__license__'],

    # Classifiers help users find your project by categorizing it.
    #
    # For a list of valid classifiers, see https://pypi.org/classifiers/
    classifiers=[  # Optional
        'Development Status :: 5 - Production/Stable',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Science/Research',

        # Pick your license as you wish
        'License :: OSI Approved :: Apache Software License',

        # Operating systems
        'Operating System :: Microsoft :: Windows :: Windows 10',
        'Operating System :: Microsoft :: Windows :: Windows 11',
        'Operating System :: MacOS',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',

        # Supported Python versions
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: Implementation :: CPython',

        # Topics
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development :: Embedded Systems',
        'Topic :: Software Development :: Testing',
        'Topic :: System :: Hardware :: Hardware Drivers',
        'Topic :: Utilities',
        'Topic :: System :: Hardware :: Universal Serial Bus (USB)',
    ],

    keywords='Joulescope',

    packages=setuptools.find_packages(exclude=['native', 'docs', 'test', 'dist', 'build']),
    ext_modules=extensions,
    cmdclass={
        'docs': CustomBuildDocs,
        'sdist': BuildSdistCommand,
    },
    include_dirs=[],
    include_package_data=True,

    # See https://packaging.python.org/guides/distributing-packages-using-setuptools/#python-requires
    python_requires='~=3.9',

    # See https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        'numpy>=1.23,<2',
        'psutil',
        "pywin32; platform_system=='Windows'",
        'requests',
    ],

    entry_points={
        'console_scripts': [
            'pyjoulescope_driver=pyjoulescope_driver.__main__:run',
        ],
    },

    project_urls={
        'Bug Reports': 'https://github.com/jetperch/pyjoulescope_driver/issues',
        'Funding': 'https://www.joulescope.com',
        'Twitter': 'https://twitter.com/joulescope',
        'Source': 'https://github.com/jetperch/pyjoulescope_driver/',
    },
)



# Copyright 2022 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Handle device programming for firmware and gateware.

For finer control on updating, use the jsdrv_util application.  For example:

jsdrv_util reset update1 && jsdrv_util mem_erase c/upd2 && jsdrv_util mem_write c/upd2 "C:/joulescope/js220/images/ctrl_release/updater2/js220_ctrl_updater2_1_0_1.img" && jsdrv_util reset update2
jsdrv_util reset update2 && jsdrv_util mem_erase c/upd1 && jsdrv_util mem_write c/upd1 "C:/joulescope/js220/images/ctrl_release/updater1/js220_ctrl_updater1_1_0_1.img" && jsdrv_util reset update1
jsdrv_util reset update1 && jsdrv_util mem_erase c/app && jsdrv_util mem_write c/app "C:/joulescope/js220/images/ctrl_release/app/js220_ctrl_app_1_0_1.img" && jsdrv_util reset app
jsdrv_util reset update1 && jsdrv_util mem_erase s/app1 && jsdrv_util mem_write s/app1 "C:/joulescope/js220/images/fpga_release/js220_fpga_1_0_1.img" && jsdrv_util reset app
"""

import json
import logging
import os
import struct
import sys
import requests
import pkgutil


_log = logging.getLogger(__name__)
APP = 'joulescope_driver'
MATURITY = ['alpha', 'beta', 'stable']
MY_PATH = os.path.dirname(os.path.abspath(__file__))
URL = 'https://download.joulescope.com/firmware/js220/'
URL_TIMEOUT = 30.0
MAGIC_HEADER = b'joulescope_img_section\x0D\x0A \x0A \x1A  \xB2\x1C'
SUBTYPE_CTRL_UPDATER1 = 1
SUBTYPE_CTRL_UPDATER2 = 2
SUBTYPE_CTRL_APP = 3
SUBTYPE_SENSOR_FPGA = 4
CTRL_TYPE_APP = 0
CTRL_TYPE_UPD1 = 1
CTRL_TYPE_UPD2 = 2


def _targets_generate():
    d = {}
    keys = ['subtype', 'name', 'mem_region', 'target_id']
    infos = [
        # [subtype,              name,           mem_ region,     target]
        [SUBTYPE_CTRL_UPDATER2, 'ctrl_updater2', 'h/mem/c/upd2', CTRL_TYPE_UPD2],
        [SUBTYPE_CTRL_UPDATER1, 'ctrl_updater1', 'h/mem/c/upd1', CTRL_TYPE_UPD1],
        [SUBTYPE_SENSOR_FPGA,   'sensor_fpga',   'h/mem/s/app1', None],
        [SUBTYPE_CTRL_APP,      'ctrl_app',      'h/mem/c/app',  CTRL_TYPE_APP],
    ]
    for info in infos:
        d[info[0]] = dict(zip(keys, info))
    return d


TARGETS = _targets_generate()


def url_save(url, filename):
    """Save a file from the distribution.

    :param url: The relative url from the main distribution URL.
    :param filename: The local filename.
    """
    _log.info(f'url_save({url}, {filename})')
    r = requests.get(URL + url, timeout=URL_TIMEOUT)
    if r.status_code != 200:
        raise FileNotFoundError(url)
    with open(filename, 'wb') as f:
        for chunk in r:
            f.write(chunk)


def release_path():
    """Get the local program storage path.

    :return: The path for storing local data.
        This path is guaranteed to exist on successful return.
    """
    if 'win32' in sys.platform:
        from win32com.shell import shell, shellcon
        path = shell.SHGetFolderPath(0, shellcon.CSIDL_LOCAL_APPDATA, None, 0)
        path = os.path.join(path, APP)
    else:
        path = os.path.expanduser('~')
        path = os.path.join(path, '.' + APP)
    path = os.path.join(path, 'program')
    os.makedirs(path, exist_ok=True)
    return path


def dist_path():
    """Get the path for distribution images.

    :return: The distribution image path.  This path is NOT
        guaranteed to exist.
    """
    return MY_PATH


def _load_file_from_network(path, force_download=None):
    cache_path = release_path()
    fname = os.path.join(cache_path, path)
    if force_download or not os.path.isfile(fname):
        os.makedirs(os.path.dirname(fname), exist_ok=True)
        url_save(path, fname)
    with open(fname, 'rb') as f:
        return f.read()


def _load_from_network(maturity=None, force_download=None):
    path = release_path()
    index_filename = os.path.join(path, 'index.json')
    url_save('index.json', index_filename)
    with open(index_filename, 'rt') as f:
        index_file = json.load(f)

    result = {}
    for target, target_value in index_file.items():
        result[target] = {}
        for m, v in target_value.items():
            if maturity is not None and m != maturity:
                continue
            v['img'] = _load_file_from_network(v['path'], force_download=force_download)
            v['changelog'] = _load_file_from_network(v['changelog'], force_download=force_download)
            result[target][m] = v
    return result


def releases_get_from_network(force_download=None, dist_save=None):
    """Get releases from the network.

    :param force_download: Force download, even if local files exist.
    :param dist_save: Save to local dist directory.
    """
    if dist_save:
        path = dist_path()
        os.makedirs(path, exist_ok=True)
    else:
        path = release_path()
    result = _load_from_network(force_download=force_download)
    for m in MATURITY:
        dst = os.path.join(path, f'img_{m}.img')
        with open(dst, 'wb') as f:
            for target_info in TARGETS.values():
                name = target_info['name']
                img = result[name][m]['img']
                subtype = struct.unpack('<H', img[296:298])[0]
                if target_info['subtype'] != subtype:
                    raise RuntimeError('subtype mismatch')
                f.write(img)
    return result


def release_get(maturity, force_download=None):
    """Get the available release.

    :param maturity: The desired maturity level, which is one of
        * alpha
        * beta
        * stable
    :param force_download: True to force download, even if local files
        are already available.  None (default) or False to use local
        files if present.
    :return: The release image as bytes
    :raise Exception: If could not load the release image.
    """
    src = None
    if maturity is None:
        maturity = 'stable'
    maturity = maturity.lower()
    if maturity not in MATURITY:
        raise ValueError(f'invalid maturity level {maturity} not in {MATURITY}')
    fname = f'img_{maturity}.img'
    rpath = os.path.join(release_path(), fname)
    if not force_download:
        try:
            img = pkgutil.get_data('pyjoulescope_driver', fname)
            _log.info('release_get found %s using pkgutil.get_data', fname)
            return img
        except FileNotFoundError:
            pass
        if os.path.isfile(rpath):
            _log.info('release_get found %s in release path', fname)
            src = rpath
    if src is None:
        try:
            releases_get_from_network(force_download=True)
            if os.path.isfile(rpath):
                _log.info('release_get downloaded %s from network', fname)
        except Exception:
            _log.warning('releases_get_from_network failed')
            raise
        src = rpath
    try:
        with open(src, 'rb') as f:
            return f.read()
    except Exception:
        _log.warning('release_get could not read %s', fname)
        raise


def release_to_segments(img: bytes):
    """Convert the binary release image into segments.

    :param img: The binary release image.
    :return: Map of segments.
    """
    segments = {}
    while len(img):
        if not img.startswith(MAGIC_HEADER):
            raise RuntimeError('invalid image')
        if len(img) < 1024:
            raise RuntimeError('invalid image')
        sz = struct.unpack('<I', img[32:36])[0]
        ver = struct.unpack('<I', img[288:292])[0]
        subtype = struct.unpack('<H', img[296:298])[0]
        segments[subtype] = {
            'subtype': subtype,
            'name': TARGETS[subtype]['name'],
            'version': ver,
            'img': img[:sz]
        }
        img = img[sz:]
    return segments


def release_to_available(img: bytes):
    """Convert the binary release image into metadata.

    :param img: The binary release image.
    :return: The available version dict[name, version]
        for each segment as a 32-bit integer
        with major8.minor8.patch16.
        The segment names are:
        * app
        * updater1
        * updater2
        * fpga
    """
    r = {}
    segments = release_to_segments(img)
    for value in segments.values():
        name = value['name'].split('_')[1]
        r[name] = value['version']
    return r


if __name__ == '__main__':
    releases_get_from_network(force_download=True, dist_save=True)
